/*
/*
 *	Low-Level PCI Support for PC
 *
 *	(c) 1999--2000 Martin Mares <mj@suse.cz>
 */


#include <sys/types.h>
#include <asm/io.h>

#include <pci.h>

void display(char []);
void printint(unsigned int);
static const struct pci_ops *conf;

struct pci_ops {
	int (*read_byte)(u8 bus, int devfn, int where, u8 *val);
	int (*read_word)(u8 bus, int devfn, int where, u16 *val);
	int (*read_dword)(u8 bus, int devfn, int where, u32 *val);
	int (*write_byte)(u8 bus, int devfn, int where, u8 val);
	int (*write_word)(u8 bus, int devfn, int where, u16 val);
	int (*write_dword)(u8 bus, int devfn, int where, u32 val);
};


/*
 * Direct access to PCI hardware...
 */


/*
 * Functions for accessing PCI configuration space with type 1 accesses
 */

#define CONFIG_CMD(bus,devfn, where)   (0x80000000 | (bus << 16) | (devfn << 8) | (where & ~3))

static int pci_conf1_read_config_byte(unsigned char bus, int devfn, int where, u8 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	*value = inb(0xCFC + (where&3));
	return 0;
}

static int pci_conf1_read_config_word(unsigned char bus, int devfn, int where, u16 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);    
	*value = inw(0xCFC + (where&2));
	return 0;    
}

static int pci_conf1_read_config_dword(unsigned char bus, int devfn, int where, u32 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	*value = inl(0xCFC);
	return 0;    
}

static int pci_conf1_write_config_byte(unsigned char bus, int devfn, int where, u8 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);    
	outb(value, 0xCFC + (where&3));
	return 0;
}

static int pci_conf1_write_config_word(unsigned char bus, int devfn, int where, u16 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	outw(value, 0xCFC + (where&2));
	return 0;
}

static int pci_conf1_write_config_dword(unsigned char bus, int devfn, int where, u32 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	outl(value, 0xCFC);
	return 0;
}

#undef CONFIG_CMD

const struct pci_ops pci_direct_conf1 = {
	pci_conf1_read_config_byte,
	pci_conf1_read_config_word,
	pci_conf1_read_config_dword,
	pci_conf1_write_config_byte,
	pci_conf1_write_config_word,
	pci_conf1_write_config_dword
};

/*
 * Functions for accessing PCI configuration space with type 2 accesses
 */

#define IOADDR(devfn, where)	((0xC000 | ((devfn & 0x78) << 5)) + where)
#define FUNC(devfn)		(((devfn & 7) << 1) | 0xf0)
#define SET(bus,devfn)		if (devfn & 0x80) return -1;outb(FUNC(devfn), 0xCF8); outb(bus, 0xCFA);

static int pci_conf2_read_config_byte(unsigned char bus, int devfn, int where, u8 *value)
{
	SET(bus, devfn);
	*value = inb(IOADDR(devfn,where));
	outb (0, 0xCF8);
	return 0;
}

static int pci_conf2_read_config_word(unsigned char bus, int devfn, int where, u16 *value)
{
	SET(bus, devfn);
	*value = inw(IOADDR(devfn,where));
	outb (0, 0xCF8);
	return 0;
}

static int pci_conf2_read_config_dword(unsigned char bus, int devfn, int where, u32 *value)
{
	SET(bus, devfn);
	*value = inl (IOADDR(devfn,where));    
	outb (0, 0xCF8);    
	return 0;
}

static int pci_conf2_write_config_byte(unsigned char bus, int devfn, int where, u8 value)
{
	SET(bus, devfn);
	outb (value, IOADDR(devfn,where));
	outb (0, 0xCF8);    
	return 0;
}

static int pci_conf2_write_config_word(unsigned char bus, int devfn, int where, u16 value)
{
	SET(bus, devfn);
	outw (value, IOADDR(devfn,where));
	outb (0, 0xCF8);    
	return 0;
}

static int pci_conf2_write_config_dword(unsigned char bus, int devfn, int where, u32 value)
{
	SET(bus, devfn);
	outl (value, IOADDR(devfn,where));    
	outb (0, 0xCF8);    
	return 0;
}

#undef SET
#undef IOADDR
#undef FUNC

const struct pci_ops pci_direct_conf2 = {
	pci_conf2_read_config_byte,
	pci_conf2_read_config_word,
	pci_conf2_read_config_dword,
	pci_conf2_write_config_byte,
	pci_conf2_write_config_word,
	pci_conf2_write_config_dword
};

/*
 * Before we decide to use direct hardware access mechanisms, we try to do some
 * trivial checks to ensure it at least _seems_ to be working -- we just test
 * whether bus 00 contains a host bridge (this is similar to checking
 * techniques used in XFree86, but ours should be more reliable since we
 * attempt to make use of direct access hints provided by the PCI BIOS).
 *
 * This should be close to trivial, but it isn't, because there are buggy
 * chipsets (yes, you guessed it, by Intel and Compaq) that have no class ID.
 */
static int pci_sanity_check(const struct pci_ops *o)
{
	u16 x;
	u8 bus;
	int devfn;
#define  PCI_CLASS_BRIDGE_HOST		0x0600
#define PCI_CLASS_DISPLAY_VGA		0x0300
#define PCI_VENDOR_ID_COMPAQ		0x0e11
#define PCI_VENDOR_ID_INTEL		0x8086

	for(devfn=0; devfn < 0x100; devfn++)
		if ((!o->read_word(bus,devfn, PCI_CLASS_SUB_CODE, &x) &&
		     (x == PCI_CLASS_BRIDGE_HOST || x == PCI_CLASS_DISPLAY_VGA)) ||
		    (!o->read_word(bus, devfn, PCI_VENDOR_ID, &x) &&
		     (x == PCI_VENDOR_ID_INTEL || x == PCI_VENDOR_ID_COMPAQ)))
			return 1;
	display("PCI: Sanity check failed\n");
	return 0;
}

static const struct pci_ops * pci_check_direct(void)
{
	unsigned int tmp;
	unsigned long flags;

	/*
	 * Check if configuration type 1 works.
	 */
	{
		outb (0x01, 0xCFB);
		tmp = inl (0xCF8);
		outl (0x80000000, 0xCF8);
		if (inl (0xCF8) == 0x80000000 &&
		    pci_sanity_check(&pci_direct_conf1)) {
			outl (tmp, 0xCF8);
			__restore_flags(flags);
			display("PCI: Using configuration type 1\n");
			return &pci_direct_conf1;
		}
		outl (tmp, 0xCF8);
	}

	/*
	 * Check if configuration type 2 works.
	 */
	{
		outb (0x00, 0xCFB);
		outb (0x00, 0xCF8);
		outb (0x00, 0xCFA);
		if (inb (0xCF8) == 0x00 && inb (0xCFA) == 0x00 &&
		    pci_sanity_check(&pci_direct_conf2)) {
			__restore_flags(flags);
			display("PCI: Using configuration type 2\n");
			return &pci_direct_conf2;
		}
	}

	return 0;
}

/*
 *  GNUBIOS
 *  Copyright 1999, 2000 Niklas Ekstrm
 *
 *  PCI Implementation
 */



static u32 NextMemAddr = 0x80000000;		/* 2GB border */
static u32 NextIoAddr = 0x1000;		/* Above all legacy IO */
static u8 MaxBus = 0;		/* The highest PCI bus number */


u32 NewPciMem(u32 size)
{
  /* well, bridges appear to like 64K alignment. Make it easy, just do that. 
   * it's also good for VM-aware OSes: you get a free red-zone for all the
   * little registers
   */
#define ROUNDPCI (64*1024-1)
  u32 NewAddr;
  NewAddr = NextMemAddr;
  NextMemAddr = (NextMemAddr + size + ROUNDPCI) & (~ROUNDPCI);
  display("NewPciMem: alloc");
  printint(NextMemAddr);
  display("\n");
  return NewAddr;
}

u32 NewPciIo(u32 size)
{
  u32 NewIoAddr = NextIoAddr;
  NextIoAddr += size;
  display("NewPciMem: alloc");
  printint(NextIoAddr);
  display("\n");

  return NewIoAddr;
}

u32
PciReadDword(u8 bus, int devfn, int reg) 
{
  u32 val;
  conf->read_dword(bus, devfn, reg, &val);
  return val;
}

u16
PciReadWord(u8 bus, int devfn, int reg) 
{
  u16 val;
  conf->read_word(bus, devfn, reg, &val);
  return val;
}

u8
PciReadByte(u8 bus, int devfn, int reg) 
{
  u8 val;
  conf->read_byte(bus, devfn, reg, &val);
  return val;
}

void
PciWriteDword(u8 bus, int devfn, int reg, u32 val)
{
  conf->write_dword(bus, devfn, reg, val);
}

void
PciWriteWord(u8 bus, int devfn, int reg, u16 val)
{
  conf->write_word(bus, devfn, reg, val);
}

void
PciWriteByte(u8 bus, int devfn, int reg, u8 val)
{
  conf->write_byte(bus, devfn, reg, val);
}

/* what we have to do: 
 * increment bus #, and set it into the device. 
 * track the current base io and memory addresses
 * iterate over the bus 
 * get sizing stuff stuff and stuff it into the bride
 * stuff last bus # into the subordinate #
 * return ...
 */
void
PciAllocateResourcesBridge(bus, devfn)
{
  int PciAllocateResourcesBus( u8 bus );
  u32 BaseMemAddr = NextMemAddr;
  u32 BaseIoAddr = NextIoAddr;

  bus;
  MaxBus = bus + 1;
  /* now set this into the starting bus # register. */
  /* hard to believe we need this ... */
  PciWriteByte(bus, devfn, PCI_PRIMARY_BUS, bus);
  PciWriteByte(bus, devfn, PCI_SECONDARY_BUS, MaxBus);
  
  /* now scan. This could recurse ... */
  PciAllocateResourcesBus(MaxBus);
  /* end of scan, set maxbus into the subordinate */
  PciWriteByte(bus, devfn, PCI_SUBORDINATE_BUS, MaxBus);
  /* set iosize and address */
  /* we need a PCI manual. Should we (I assume) set a bit in the 
   * cmd register if I/O will work? I'm unsure about bridges. -- RGM
   */
  PciWriteWord(bus, devfn, PCI_IO_BASE, BaseIoAddr);
  if (NextIoAddr > BaseIoAddr)
    PciWriteWord(bus, devfn, PCI_IO_BASE, NextIoAddr-1);
  else
    PciWriteWord(bus, devfn, PCI_IO_BASE, BaseIoAddr);

  PciWriteWord(bus, devfn, PCI_MEMORY_BASE, BaseMemAddr);
  PciWriteWord(bus, devfn, PCI_MEMORY_LIMIT, NextIoAddr);

}
int PciAllocateResourcesBus( u8 bus )
{
	int devfn;
	u16 command;
	u8 reg;
	u32 addr;
	u8 membits;

	for( devfn = 0; devfn < 256; devfn++ )
	{
	  u8 type;
		if( PciReadDword( bus, devfn, PCI_VENDOR_ID ) == 0xffffffff )
		{
			devfn |= 0x07;
			continue;
		}

		type = PciReadByte( bus, devfn, PCI_HEADER_TYPE );
		type &= 0x7f;
		if (type == PCI_HEADER_TYPE_P2P)
		  PciAllocateResourcesBridge(bus, devfn);
		/* other than that, it better be 0 ... */
		if( type )
		  continue;

		command = 0;

		for( reg = PCI_BASE_ADDRESS_0; reg <= PCI_BASE_ADDRESS_5; reg += 4 )
		{
			PciWriteDword( bus, devfn, reg, 0xffffffff );
			addr = PciReadDword( bus, devfn, reg );
			if( addr == 0 )
				continue;

			if( (addr & PCI_ADDR_SPACE) == PCI_ADDR_SPACE_MEM )			/* MEM */
			{
				command |= PCI_COMMAND_MEMORY;

				membits = addr & 0xf;
				addr &= 0xfffffff0;
				NextMemAddr = (NextMemAddr + ~addr) & addr;
				PciWriteDword( bus, devfn, reg, NextMemAddr | membits );
				display("mallocate "); printint(NextMemAddr);
				display("\n");
				NextMemAddr += ~addr + 1;
			}
			else																										/* IO */
			{
				command |= PCI_COMMAND_IO;

				addr &= 0xfffffffc;
				addr |= 0xffff0000;					// The IO address space is only 16 bits
				NextIoAddr = (NextIoAddr + ~addr) & addr;
				display("iallocate "); printint(NextMemAddr);
				display("\n");

				PciWriteDword( bus, devfn, reg, NextIoAddr | PCI_ADDR_SPACE_IO );

				NextIoAddr += ~addr + 1;
			}
		}

		PciWriteWord( bus, devfn, PCI_COMMAND, command );

		PciWriteDword( bus, devfn, PCI_ROM_BASE_ADDRESS, 0xffffffff );
		addr = PciReadDword( bus, devfn, PCI_ROM_BASE_ADDRESS );
		if( addr != 0 )
		{
			addr &= 0xfffff800;
			NextMemAddr = (NextMemAddr + ~addr) & addr;
			PciWriteDword( bus, devfn, PCI_ROM_BASE_ADDRESS, NextMemAddr | PCI_ROM_BASE_ENABLED );

			NextMemAddr += ~addr + 1;
		}
	}
}

int PciAllocateResources()
{
  conf = &pci_direct_conf1;
  conf = pci_check_direct();
  pci_sanity_check(conf);
  return( PciAllocateResourcesBus( 0 ) );
}

#if 0
int PciFindDevice( u8 bus, u32 venddev, int devfn )
{
	u32 tmp;

	for ( ; devfn < 256; devfn++ )
	{
		tmp = PciReadDword( bus, devfn, PCI_VENDOR_ID );
		if( tmp == 0xffffffff )
			devfn |= 0x07;
		else if( tmp == venddev )
			return( devfn );
	}

	return( -1 );
}

int PciFindClass( u8 bus, u32 devclass, int devfn )
{
	devclass <<= 8;

	for ( ; devfn < 256; devfn++ )
	{
		if( PciReadDword( bus, devfn, PCI_VENDOR_ID ) == 0xffffffff )
			devfn |= 0x07;
		else if( (PciReadDword( bus, devfn, PCI_CLASS_REVISION ) & PCI_CLASS_MASK) == devclass )
			return( devfn );
	}

	return( -1 );
}
#endif
 *	Low-Level PCI Support for PC
 *
 *	(c) 1999--2000 Martin Mares <mj@suse.cz>
 */

#include <sys/types.h>
#include <asm/io.h>


#include "pci-i386.h"

int pcibios_last_bus = -1;

/*
 * Direct access to PCI hardware...
 */


/*
 * Functions for accessing PCI configuration space with type 1 accesses
 */

#define CONFIG_CMD(bus,devfn, where)   (0x80000000 | (bus << 16) | (devfn << 8) | (where & ~3))

static int pci_conf1_read_config_byte(unsigned char bus, int devfn, int where, u8 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	*value = inb(0xCFC + (where&3));
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf1_read_config_word(unsigned char bus, int devfn, int where, u16 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);    
	*value = inw(0xCFC + (where&2));
	return PCIBIOS_SUCCESSFUL;    
}

static int pci_conf1_read_config_dword(unsigned char bus, int devfn, int where, u32 *value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	*value = inl(0xCFC);
	return PCIBIOS_SUCCESSFUL;    
}

static int pci_conf1_write_config_byte(unsigned char bus, int devfn, int where, u8 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);    
	outb(value, 0xCFC + (where&3));
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf1_write_config_word(unsigned char bus, int devfn, int where, u16 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	outw(value, 0xCFC + (where&2));
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf1_write_config_dword(unsigned char bus, int devfn, int where, u32 value)
{
	outl(CONFIG_CMD(bus,devfn,where), 0xCF8);
	outl(value, 0xCFC);
	return PCIBIOS_SUCCESSFUL;
}

#undef CONFIG_CMD

static struct pci_ops pci_direct_conf1 = {
	pci_conf1_read_config_byte,
	pci_conf1_read_config_word,
	pci_conf1_read_config_dword,
	pci_conf1_write_config_byte,
	pci_conf1_write_config_word,
	pci_conf1_write_config_dword
};

/*
 * Functions for accessing PCI configuration space with type 2 accesses
 */

#define IOADDR(devfn, where)	((0xC000 | ((devfn & 0x78) << 5)) + where)
#define FUNC(devfn)		(((devfn & 7) << 1) | 0xf0)
#define SET(dev)		if (dev->devfn & 0x80) return PCIBIOS_DEVICE_NOT_FOUND;		\
				outb(FUNC(dev->devfn), 0xCF8);					\
				outb(dev->bus->number, 0xCFA);

static int pci_conf2_read_config_byte(unsigned char bus, int devfn, int where, u8 *value)
{
	SET(dev);
	*value = inb(IOADDR(dev->devfn,where));
	outb (0, 0xCF8);
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf2_read_config_word(unsigned char bus, int devfn, int where, u16 *value)
{
	SET(dev);
	*value = inw(IOADDR(dev->devfn,where));
	outb (0, 0xCF8);
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf2_read_config_dword(unsigned char bus, int devfn, int where, u32 *value)
{
	SET(dev);
	*value = inl (IOADDR(dev->devfn,where));    
	outb (0, 0xCF8);    
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf2_write_config_byte(unsigned char bus, int devfn, int where, u8 value)
{
	SET(dev);
	outb (value, IOADDR(dev->devfn,where));
	outb (0, 0xCF8);    
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf2_write_config_word(unsigned char bus, int devfn, int where, u16 value)
{
	SET(dev);
	outw (value, IOADDR(dev->devfn,where));
	outb (0, 0xCF8);    
	return PCIBIOS_SUCCESSFUL;
}

static int pci_conf2_write_config_dword(unsigned char bus, int devfn, int where, u32 value)
{
	SET(dev);
	outl (value, IOADDR(dev->devfn,where));    
	outb (0, 0xCF8);    
	return PCIBIOS_SUCCESSFUL;
}

#undef SET
#undef IOADDR
#undef FUNC

static struct pci_ops pci_direct_conf2 = {
	pci_conf2_read_config_byte,
	pci_conf2_read_config_word,
	pci_conf2_read_config_dword,
	pci_conf2_write_config_byte,
	pci_conf2_write_config_word,
	pci_conf2_write_config_dword
};

/*
 * Before we decide to use direct hardware access mechanisms, we try to do some
 * trivial checks to ensure it at least _seems_ to be working -- we just test
 * whether bus 00 contains a host bridge (this is similar to checking
 * techniques used in XFree86, but ours should be more reliable since we
 * attempt to make use of direct access hints provided by the PCI BIOS).
 *
 * This should be close to trivial, but it isn't, because there are buggy
 * chipsets (yes, you guessed it, by Intel and Compaq) that have no class ID.
 */
static int __init pci_sanity_check(struct pci_ops *o)
{
	u16 x;
	struct pci_bus bus;		/* Fake bus and device */
	struct pci_dev dev;

	if (pci_probe & PCI_NO_CHECKS)
		return 1;
	bus.number = 0;
	dev.bus = &bus;
	for(dev.devfn=0; dev.devfn < 0x100; dev.devfn++)
		if ((!o->read_word(&dev, PCI_CLASS_DEVICE, &x) &&
		     (x == PCI_CLASS_BRIDGE_HOST || x == PCI_CLASS_DISPLAY_VGA)) ||
		    (!o->read_word(&dev, PCI_VENDOR_ID, &x) &&
		     (x == PCI_VENDOR_ID_INTEL || x == PCI_VENDOR_ID_COMPAQ)))
			return 1;
	DBG("PCI: Sanity check failed\n");
	return 0;
}

static struct pci_ops * __init pci_check_direct(void)
{
	unsigned int tmp;
	unsigned long flags;

	/*
	 * Check if configuration type 1 works.
	 */
	{
		outb (0x01, 0xCFB);
		tmp = inl (0xCF8);
		outl (0x80000000, 0xCF8);
		if (inl (0xCF8) == 0x80000000 &&
		    pci_sanity_check(&pci_direct_conf1)) {
			outl (tmp, 0xCF8);
			__restore_flags(flags);
			printk("PCI: Using configuration type 1\n");
			return &pci_direct_conf1;
		}
		outl (tmp, 0xCF8);
	}

	/*
	 * Check if configuration type 2 works.
	 */
	{
		outb (0x00, 0xCFB);
		outb (0x00, 0xCF8);
		outb (0x00, 0xCFA);
		if (inb (0xCF8) == 0x00 && inb (0xCFA) == 0x00 &&
		    pci_sanity_check(&pci_direct_conf2)) {
			__restore_flags(flags);
			printk("PCI: Using configuration type 2\n");
			return &pci_direct_conf2;
		}
	}

	return NULL;
}
