#ifndef ROM_DEVICE_H
#define	ROM_DEVICE_H

/*
 * Bootstrap code for the STPC Consumer
 * Copyright (c) 1999 by Net Insight AB. All Rights Reserved.
 *
 * $Id: device.h,v 1.2 2000/03/05 21:52:13 jgarzik Exp $
 *
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <rom/queue.h>

/*
 *	Max length of a devince name (plus unit number).
 */
#define	SZDEVNAME	8

/** forward prototype */
struct device_unit;

/*
 *	Device operators. Devices that can be handled (and are not
 *	just state devices) uses these callbacks.
 */
typedef struct device_ops {
	struct device_unit *(*open) __P((struct device_unit *));		/* open device */
	int		(*close) __P((struct device_unit *));			/* close device */
	int		(*read) __P((struct device_unit *, char *, int));	/* read from dev */
	int		(*write) __P((struct device_unit *, char *, int));	/* write to dev */
	int		(*ioctl) __P((struct device_unit *, int, void *));	/* ioctl device */
} *device_ops_t;
#define	DEVICE_OPS_NULL	((device_ops_t) 0)

/*
 *	This is the device structure. It is created by the autoconfiguration
 *	sequence and is passed on to the device attach routine.
 */
typedef struct device {
	queue_chain_t	dv_link;		/* link node */
	struct cfdata *	dv_cfdata;		/* config data */
	char *		dv_softc;		/* softc */
	u_int32_t	dv_unit;		/* unit */
	struct device *	dv_parent;		/* parent device */
	u_int32_t	dv_flags;		/* flags */
	char		dv_name[SZDEVNAME];	/* name */
	device_ops_t	dv_ops;			/* operations */
} *device_t;
#define	DEVICE_NULL	((device_t) 0)

/*
 *	Unit structure. All the I/O routines uses this structure to 
 *	control I/O.
 */
typedef struct device_unit {
	device_t	du_device;		/* device */
	int		du_unit;		/* unit */
	u_int32_t	du_offset;		/* offset */
	void *		du_data;		/* private data */
} *device_unit_t;
#define	DEVICE_UNIT_NULL ((device_unit_t) 0)

/*
 *	Routine prototypes that the autoconfiguration sequence uses to find
 *	the right device for the specified hardware.
 */
typedef	int (*cfmatch_t) __P((device_t, struct cfdata *, void *));

/*
 *	Configuration hooks for a device. 
 */
typedef struct cfattach {
	u_int32_t	ca_size;		/* sizeof device data */
	cfmatch_t	ca_match;
	int32_t		(*ca_attach) __P((device_t, device_t, void *));	
	int32_t		(*ca_detach) __P((device_t));
	char *		ca_name;		/* device name */
} *cfattach_t;
#define	CFATTACH_NULL	((cfattach_t) 0)

/*
 *	Configuration driver structure. Contains information about
 *	a driver.
 */
typedef struct cfdriver {
	unsigned int	cd_nunits;	/* number of units */
} *cfdriver_t;
#define	CFDRIVER_NULL	((cfdriver_t) 0)

/*
 *	Configuration data.
 */
typedef struct cfdata {
	cfattach_t	cf_attach;	/* attach */
	struct cfdriver	*cf_driver;	/* driver */
	int		cf_unit;	/* unit */	
	char *		cf_data;	/* driver data */
	unsigned int	cf_flags;	/* flags */
	struct cfdata	*cf_parent;	/* parent device */
} *cfdata_t;
#define	CFDATA_NULL	((cfdata_t) 0)

/** flags: */
#define	CF_FLAG_NOTFOUND	0	/* device not found yet */
#define	CF_FLAG_FOUND		1	/* device found */
#define	CF_FLAG_STAR		2	/* wildcard device, leave untouched */

#define	CF_MATCH_NONE		0
#define	CF_MATCH_LOW		1
#define	CF_MATCH_GOOD		2
#define	CF_MATCH_EXACT		3

/*
 *	I/O request structure:
 */
typedef struct iorequest {
	device_unit_t		ior_unit;	/* unit */
	unsigned int		ior_action;	/* action */
	char *			ior_buffer;	/* buffer */
	unsigned int		ior_length;	/* length */
	unsigned int		ior_actual;	/* actual length */
} *iorequest_t;
#define	IOREQUEST_NULL	((iorequest_t) 0)

/*
 *	Block I/O request structure:
 *
 *	Used to read / write to block devices.
 */
typedef struct block_iorequest {
	struct iorequest	bir_ioreq;	/* std i/o request */
	unsigned int		bir_block;	/* block # to read */
	unsigned int		bir_nblocks;	/* number of blocks to read */
} *block_iorequest_t;
#define	BLOCK_IOREQUEST_NULL	((block_iorequest_t) 0)

/*
 *	I/O requests:
 */
#define	IOR_OPEN		0
#define	IOR_CLOSE		1
#define	IOR_READ		2
#define	IOR_WRITE		3
#define	IOR_IOCTL		4
#define	IOR_READBLOCK		5
#define	IOR_WRITEBLOCK		6

/*
 *	Prototypes:
 */
extern int device_blk_read(
		device_unit_t,	/* unit */
		u_int32_t,	/* sector */
		u_int32_t,	/* nsectors */
		char *);	/* buffer */

#endif /* ROM_DEVICE_H */
